/*
 * Joe Meadows' UNMESSAGE Utility (decompile messages)
 * available from http://vms.process.com/scripts/fileserv/fileserv.com
 * 
 * Modified by Hartmut Becker, Dec 2010: simplified to make it work for
 * nonexecutable message files from all VMS platforms.
 * 
 * Build:   $ cc/standard=relaxed unmsg
 *          $ link unmsg
 * Usage:   $ mc sys$disk:[]unmsg <message_file>
 * Example: $ mc sys$disk:[]unmsg sys$message:sysmsg.exe
 */
#include <stdio.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>

#include "unmsg.h"

#define $VMS_STATUS_FAC_NO(s) (((s)&0xfff0000)>>16)
#define $VMS_STATUS_SEVERITY(s) ((s)&0x7)
#define $VMS_STATUS_CODE(s) (((s)&0x7ff8)>>3)

struct Ascic
{
  unsigned char len;
  char text[];
};

static int decompile_msc(struct mscdef *msc);
static int decompile_midx(struct mscdef *msc,
                   struct midx$r_entry *entry);
static int print_message(struct mscdef *msc,
              struct midx$r_entry *entry);
static char *ifquote(struct Ascic *s);

int main(int argc, char *argv[])
{
  int status;
  int fd;
  int *map_adr;
  struct stat buf;

  fd = open (argv[1], O_RDONLY);
  if (fd==-1)
     return perror("open failed"), EXIT_FAILURE;
  if (-1==fstat(fd, &buf))
     return perror("fstat failed"), EXIT_FAILURE;

  map_adr = mmap(NULL, buf.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
  if (map_adr==MAP_FAILED)
     return perror("mmap failed"), EXIT_FAILURE;

  map_adr = map_adr + (512/sizeof(int));
  map_adr = map_adr + (16/sizeof(int));
  while(*map_adr)
  {
    status = decompile_msc((struct mscdef*)((int)map_adr + *map_adr));
    if (!status) return EXIT_FAILURE;

    ++map_adr; /* point to the next int .. */
  }
  return EXIT_SUCCESS;
}

static int decompile_msc(struct mscdef *msc)
{
  struct midxdef *midx;
  struct midx$r_entry *entry;
  int status;
  int i;

  if (msc->msc$w_sanity != MSC$C_SANITY)
    return 0;

  midx = (struct midxdef *)((int)msc + msc->msc$l_index_off);
  if (midx->midx$b_sanity != MIDX$C_SANITY)
    return 0;

  entry = (struct midx$r_entry *)((int)midx + MIDX$C_LENGTH);

  for (i = 1; i < (midx->midx$w_size / MIDX$C_LENGTH); ++i)
  {
    status = decompile_midx(msc, entry);
    if (!status) return status;

    ++entry; /* skip to next entry.. */
  }

  return 1;
}

static int decompile_midx(struct mscdef *msc,
                   struct midx$r_entry *entry)
{
  struct midxdef *midx;
  int status;
  int i;
  
  if ((int)(entry->midx$l_offset) & 1)
  { /* this is a subindex */
    midx = (struct midxdef *)((int)msc + (int)entry->midx$l_offset - 1);

    entry = (struct midx$r_entry *)((int)midx + MIDX$C_LENGTH);

    for (i = 1; i < (midx->midx$w_size / MIDX$C_LENGTH); ++i)
    {
      status = decompile_midx(msc, entry);
      if (!status) return status;

      ++entry;
    }
    return 1;
  }
  else
  {
    return print_message(msc, entry);
  }
}

static int print_message(struct mscdef *msc,
              struct midx$r_entry *entry)
{
  struct mrecdef *mrec = (struct mrecdef *)((int)msc + entry->midx$l_offset);
  static int facility = -1;
  static int base = -1;
  static int severity = -1;
  struct Ascic *msg_ascic;

  if ($VMS_STATUS_FAC_NO(entry->midx$l_code) != facility)
  {
    short int *mfac_size = (short int *)((int)msc + msc->msc$l_fac_off);
    struct mfacdef
      *mfac = (struct mfacdef *)((int)msc + msc->msc$l_fac_off + 2),
      *mfac_end = (struct mfacdef *)((int)mfac_size + *mfac_size);

    facility = $VMS_STATUS_FAC_NO(entry->midx$l_code);
    while((mfac < mfac_end) && (mfac->mfac$w_number != facility))
    {
      mfac = (struct mfacdef *)((int)mfac + mfac->mfac$b_namelen + 3);
    }
    if (mfac == mfac_end)
      return 0;

    printf(".FACILITY\t%.*s,%d%s%s",
           mfac->mfac$b_namelen, mfac->mfac$t_name,
           mfac->mfac$w_number & 0x07ff,
           ((mfac->mfac$w_number & 0x8000)
           || (mfac->mfac$w_number == 0))?" /SHARED":"",
           ((mfac->mfac$w_number & 0x0800) == 0)?" /SYSTEM":"");
    printf ("\n");

    base = 1;
    severity = -1;
  }

  if ($VMS_STATUS_SEVERITY(entry->midx$l_code) != severity)
  {
    static char *severities[] =
      { "WARNING", "SUCCESS", "ERROR", "INFORMATIONAL", "FATAL", "?" };

    severity = $VMS_STATUS_SEVERITY(entry->midx$l_code);

    printf(".SEVERITY\t%s", severities[severity]);
    printf ("\n");

  }

  if ($VMS_STATUS_CODE(entry->midx$l_code) != base)
  {
    base = $VMS_STATUS_CODE(entry->midx$l_code);
    printf(".BASE\t\t%d", base);
    printf ("\n");

  }

  ++base;

  msg_ascic = (struct Ascic *)
              (&mrec->mrec$b_identlen + mrec->mrec$b_identlen + 1);
  
  if (mrec->mrec$b_faocnt)
  {
    char *quoted = ifquote(msg_ascic);

    printf("_%08X\t%s%.*s%s /FAO=%d /IDENT=%.*s",
                  entry->midx$l_code,
                  quoted?"<":"\"",
                  msg_ascic->len, msg_ascic->text,
                  quoted?">":"\"",
                  mrec->mrec$b_faocnt,
                  mrec->mrec$b_identlen, mrec->mrec$t_ident);
    printf ("\n");
  }
  else
  {
    char *quoted = ifquote(msg_ascic);

    printf("_%08X\t%s%.*s%s /IDENT=%.*s",
                  entry->midx$l_code,
                  quoted?"<":"\"",
                  msg_ascic->len, msg_ascic->text,
                  quoted?">":"\"",
                  mrec->mrec$b_identlen, mrec->mrec$t_ident);
    printf ("\n");
  }
  return 1;
}

static char *ifquote(struct Ascic *s)
{
  int i;

  for (i = 0; i < s->len; ++i)
  {
    if (s->text[i] == '\"') return &(s->text[i]);
  }

  return NULL;
}
